-- complain if script is sourced in psql, rather than via CREATE EXTENSION
\echo Use "CREATE EXTENSION timeit" to load this file. \quit
CREATE TYPE timeit.measure_type AS ENUM ('cycles', 'time');
DO $$
BEGIN
    -- Check if the 'trim_scale' function exists in the 'pg_catalog' schema
    IF NOT EXISTS (
        SELECT 1
        FROM pg_proc
        WHERE proname = 'trim_scale'
        AND pronamespace = (SELECT oid FROM pg_namespace WHERE nspname = 'pg_catalog')
    ) THEN
        -- Create the 'trim_scale' function if it doesn't exist
        EXECUTE
        $_$
            CREATE OR REPLACE FUNCTION timeit.trim_scale(numeric) RETURNS numeric AS
            $__$
                BEGIN
                    -- Convert the numeric to text, trim trailing zeros and then convert back to numeric
                    RETURN regexp_replace($1::text, '(\.\d*?[1-9])0+$|\.0+$', '\1')::numeric;
                END;
            $__$ LANGUAGE plpgsql IMMUTABLE;
        $_$;
    END IF;
END $$;
CREATE OR REPLACE FUNCTION timeit.round_to_sig_figs(numeric, integer)
RETURNS numeric
LANGUAGE sql
AS $$
SELECT round($1, $2 - 1 - floor(coalesce(log(nullif(abs($1),0)),0))::int);
$$;

CREATE OR REPLACE FUNCTION timeit.round_to_sig_figs(bigint, integer)
RETURNS bigint
LANGUAGE sql
AS $$
SELECT round($1, $2 - 1 - floor(coalesce(log(nullif(abs($1),0)),0))::int)::bigint;
$$;
--
-- Returns measured execution_time in human-readable output,
-- using time unit suffixes, i.e. "ns", "us", "ms".
--
CREATE OR REPLACE FUNCTION timeit.pretty_time(numeric)
RETURNS text
LANGUAGE sql
SET search_path = timeit, public, pg_temp
AS $$
SELECT
    CASE
        WHEN $1 IS NULL THEN 'NULL'
        WHEN log10(nullif(abs($1),0)) >= 0 THEN format('%s s', $1)
        WHEN log10(nullif(abs($1),0)) >= -3 THEN format('%s ms', trim_scale($1 * 1e3))
        WHEN log10(nullif(abs($1),0)) >= -6 THEN format('%s µs', trim_scale($1 * 1e6))
        WHEN log10(nullif(abs($1),0)) >= -9 THEN format('%s ns', trim_scale($1 * 1e9))
        ELSE format('%s s', $1::float8::text)
    END
$$;

CREATE OR REPLACE FUNCTION timeit.pretty_time(numeric, significant_figures integer)
RETURNS text
LANGUAGE sql
AS $$
SELECT timeit.pretty_time(timeit.round_to_sig_figs($1, $2))
$$;
CREATE OR REPLACE FUNCTION timeit.compute_regression_metrics
(
    x float8[],
    y float8[]
)
RETURNS TABLE
(
    r_squared float8,
    slope float8,
    intercept float8
)
LANGUAGE plpgsql
AS $$
declare
    sum_x float8 := 0;
    sum_y float8 := 0;
    sum_xy float8 := 0;
    sum_x2 float8 := 0;
    sum_y2 float8 := 0;
    n integer := array_length(x, 1);
begin
    if (n >= 2) is not true then
        raise exception 'input arrays must have at least two elements';
    end if;
    if n is distinct from array_length(y, 1) then
        raise exception 'input arrays must be of same length';
    end if;

    for i in 1..n loop
        sum_x := sum_x + x[i];
        sum_y := sum_y + y[i];
        sum_xy := sum_xy + x[i] * y[i];
        sum_x2 := sum_x2 + x[i]^2;
        sum_y2 := sum_y2 + y[i]^2;
    end loop;

    if (n * sum_x2 - sum_x^2) <> 0 and (n * sum_y2 - sum_y^2) <> 0 then
        r_squared := (n * sum_xy - sum_x * sum_y)^2 / ((n * sum_x2 - sum_x^2) * (n * sum_y2 - sum_y^2));
        slope := (n * sum_xy - sum_x * sum_y) / (n * sum_x2 - sum_x^2);
        intercept := (sum_y - slope * sum_x) / n;
        return next;
    end if;
end
$$;
CREATE OR REPLACE FUNCTION timeit.measure_time
(
    internal_function text,
    input_values text[],
    iterations bigint,
    core_id integer
)
RETURNS bigint
LANGUAGE c
AS '$libdir/timeit', 'measure_time';
CREATE OR REPLACE FUNCTION timeit.measure_cycles
(
    internal_function text,
    input_values text[],
    iterations bigint,
    core_id integer
)
RETURNS bigint
LANGUAGE c
AS '$libdir/timeit', 'measure_cycles';
CREATE OR REPLACE FUNCTION timeit.overhead_time
(
    internal_function text,
    input_values text[],
    iterations bigint,
    core_id integer
)
RETURNS bigint
LANGUAGE c
AS '$libdir/timeit', 'overhead_time';
CREATE OR REPLACE FUNCTION timeit.overhead_cycles
(
    internal_function text,
    input_values text[],
    iterations bigint,
    core_id integer
)
RETURNS bigint
LANGUAGE c
AS '$libdir/timeit', 'overhead_cycles';
CREATE OR REPLACE FUNCTION timeit.eval
(
    internal_function text,
    input_values text[]
)
RETURNS text
LANGUAGE c
AS '$libdir/timeit', 'eval';
CREATE OR REPLACE FUNCTION timeit.measure
(
    function_name text,
    input_values text[],
    r_squared_threshold float8 DEFAULT 0.99,
    sample_size integer DEFAULT 10,
    timeout interval DEFAULT '1 second'::interval,
    measure_type timeit.measure_type DEFAULT 'time',
    core_id integer DEFAULT -1 /* -1 means let the OS schedule CPU core */
)
RETURNS TABLE
(
    x float8[],
    y float8[],
    r_squared float8,
    slope float8,
    intercept float8,
    iterations bigint
)
LANGUAGE plpgsql
AS $$
declare
    res float8;
    n integer;
    t0 timestamptz;
    is_timeout boolean;
begin
    if (r_squared_threshold between 0.99 and 0.99999) is not true then
        raise exception 'r_squared_threshold must be between 0.99 and 0.99999';
    end if;

    iterations := 1;
    x := ARRAY[]::float8[];
    y := ARRAY[]::float8[];
    t0 := clock_timestamp();
    loop
        if measure_type = 'cycles' then
            res := timeit.measure_cycles(function_name, input_values, iterations, core_id)::float8;
        elsif measure_type = 'time' then
            res := timeit.measure_time(function_name, input_values, iterations, core_id)::float8;
        else
            raise exception 'invalid measure_type %', measure_type;
        end if;

        y := array_append(y, res);
        x := array_append(x, iterations::float8);
        n := array_length(y, 1);

        is_timeout := clock_timestamp() - t0 > timeout;

        if n >= sample_size or (n >= 2 and is_timeout) then
            SELECT r.r_squared, r.slope, r.intercept
            INTO r_squared, slope, intercept
            FROM timeit.compute_regression_metrics(x[n-sample_size+1:n], y[n-sample_size+1:n]) AS r;

            raise debug 'Predicted formula: y = % * x + %, r_squared = %, iterations = %', slope, intercept, r_squared, iterations;

            if r_squared >= r_squared_threshold and slope > 0 then
                return next;
                return;
            end if;

            if is_timeout then
                raise debug 'Timeout';
                return next;
                return;
            end if;
        end if;

        iterations := iterations * 2;
    end loop;
end
$$;
--
-- Returns measured execution time in human-readable output,
-- using time unit suffixes, i.e. "ns", "us", "ms".
--
CREATE OR REPLACE FUNCTION timeit.t
(
    function_name text,
    input_values text[],
    significant_figures integer DEFAULT 1,
    r_squared_threshold float8 DEFAULT 0.99,
    sample_size integer DEFAULT 10,
    timeout interval DEFAULT '1 second'::interval,
    core_id integer DEFAULT -1 /* -1 means let the OS schedule CPU core */
)
RETURNS text
LANGUAGE SQL AS
$$
    -- slope value is in microseconds for measure_type 'time'
    SELECT timeit.pretty_time((measure.slope/1e6)::numeric, significant_figures)
    FROM timeit.measure(function_name, input_values, r_squared_threshold,
                        sample_size, timeout, 'time', core_id);
$$;
--
-- Returns measured clock cycles as a bigint,
-- rounded to significant_figures
--
CREATE OR REPLACE FUNCTION timeit.c
(
    function_name text,
    input_values text[],
    significant_figures integer DEFAULT 1,
    r_squared_threshold float8 DEFAULT 0.99,
    sample_size integer DEFAULT 10,
    timeout interval DEFAULT '1 second'::interval,
    core_id integer DEFAULT -1 /* -1 means let the OS schedule CPU core */
)
RETURNS bigint
LANGUAGE SQL AS
$$
    SELECT timeit.round_to_sig_figs(measure.slope::bigint, significant_figures)
    FROM timeit.measure(function_name, input_values, r_squared_threshold,
                        sample_size, timeout, 'cycles', core_id);
$$;
